/**
 * dte.c
 * =====
 * Implementation of a simple Decision Tree Evaluator.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#include <stdio.h>
#include <stdlib.h>

#include "dte.h"


/**
 * Handle for a generic node in a decision tree.
 */
typedef struct node *Node;

/**
 * Handle for a decision node in a decision tree.
 */
typedef struct decision_node *DecisionNode;

/**
 * Handle for a chance node in a decision tree.
 */
typedef struct chance_node *ChanceNode;

/**
 * Handle for a terminal node in a decision tree.
 */
typedef struct terminal_node *TerminalNode;

/**
 * A decision tree.
 */
struct decision_tree {
    Node root;         /* handle for root node */
    double *values;    /* array of values with probabilities first,
                          followed by utilities */
    double *utilities; /* points to the first utility value in *values" */
};

/**
 * Node types.
 */
typedef enum { DECISION_NODE, CHANCE_NODE, TERMINAL_NODE } NodeType;

/**
 * A generic node.
 */
struct node {
    NodeType type;      /* type of node */
    union {
	DecisionNode d;
	ChanceNode c;
	TerminalNode t;
    } data;             /* node data */ 
};

/**
 * A decision node.
 */
struct decision_node {
    DecisionTree tree;  /* handle for tree that this node belongs to */
    Node *alternatives; /* handles to alternatives */
    int nalternatives;  /* number of alternatives */
};

/**
 * A chance node.
 */
struct chance_node {
    DecisionTree tree; /* handle for tree that this node belongs to */
    Node *outcomes;    /* handles to outcomes */
    int *probindexes;  /* indexes to probabilities in value array of tree */
    int noutcomes;     /* number of outcomes */
};

/**
 * A terminal node.
 */
struct terminal_node {
    DecisionTree tree; /* handle for tree that this node belongs to */
    int utilindex;     /* index to utility in value array of tree */
};


static DecisionTree createDecisionTree(void);
static void initDecisionTree(DecisionTree tree, int nprobs, int nutils);

static Node createDecisionNode(DecisionTree tree, int nalternatives);
static Node createChanceNode(DecisionTree tree, int noutcomes);
static Node createTerminalNode(DecisionTree tree);
static void freeNode(Node node);
static void freeDecisionNode(DecisionNode node);
static void freeChanceNode(ChanceNode node);
static void freeTerminalNode(TerminalNode node);
static Node readNode(FILE *file, DecisionTree tree, int *maxp, int *maxu);
static Node readDecisionNode(FILE *file, DecisionTree tree,
			     int *maxp, int *maxu);
static Node readChanceNode(FILE *file, DecisionTree tree,
			   int *maxp, int *maxu);
static Node readTerminalNode(FILE *file, DecisionTree tree,
			     int *maxp, int *maxu);
static double expectedUtility(Node node);
static double expectedUtilityD(DecisionNode node);
static double expectedUtilityC(ChanceNode node);
static double expectedUtilityT(TerminalNode node);
static int evaluateDecisionNode(DecisionNode node);


static DecisionTree createDecisionTree(void)
{
    DecisionTree tree;

    tree = (DecisionTree)malloc(sizeof(*tree));

    return tree;
}

static void initDecisionTree(DecisionTree tree, int nprobs, int nutils)
{
    tree->values = (double*)malloc((nprobs + nutils) * sizeof(*tree->values));
    tree->utilities = tree->values + nprobs;
}

static Node createDecisionNode(DecisionTree tree, int nalternatives)
{
    Node node;

    node = (Node)malloc(sizeof(*node));
    node->type = DECISION_NODE;
    node->data.d = (DecisionNode)malloc(sizeof(*node->data.d));
    node->data.d->tree = tree;
    node->data.d->alternatives =
	(Node*)malloc(nalternatives * sizeof(*node->data.d->alternatives));
    node->data.d->nalternatives = nalternatives;

    return node;
}

static Node createChanceNode(DecisionTree tree, int noutcomes)
{
    Node node;

    node = (Node)malloc(sizeof(*node));
    node->type = CHANCE_NODE;
    node->data.c = (ChanceNode)malloc(sizeof(*node->data.c));
    node->data.c->tree = tree;
    node->data.c->outcomes =
	(Node*)malloc(noutcomes * sizeof(*node->data.c->outcomes));
    node->data.c->probindexes =
	(int*)malloc(noutcomes * sizeof(*node->data.c->probindexes));
    node->data.c->noutcomes = noutcomes;

    return node;
}

static Node createTerminalNode(DecisionTree tree)
{
    Node node;

    node = (Node)malloc(sizeof(*node));
    node->type = TERMINAL_NODE;
    node->data.t = (TerminalNode)malloc(sizeof(*node->data.t));
    node->data.t->tree = tree;

    return node;
}

static void freeNode(Node node)
{
    switch (node->type)
    {
    case DECISION_NODE:
	freeDecisionNode(node->data.d);
	break;
    case CHANCE_NODE:
	freeChanceNode(node->data.c);
	break;
    case TERMINAL_NODE:
	freeTerminalNode(node->data.t);
	break;
    }
    free(node);
}

static void freeDecisionNode(DecisionNode node)
{
    int i;

    for (i = 0; i < node->nalternatives; i++)
	freeNode(node->alternatives[i]);
    free(node->alternatives);
    free(node);
}

static void freeChanceNode(ChanceNode node)
{
    int i;

    for (i = 0; i < node->noutcomes; i++)
	freeNode(node->outcomes[i]);
    free(node->outcomes);
    free(node->probindexes);
    free(node);
}

static void freeTerminalNode(TerminalNode node)
{
    free(node);
}

static Node readNode(FILE *file, DecisionTree tree, int *maxp, int *maxu)
{
    char type;

    fscanf(file, "%*[\n\t ]");
    fscanf(file, "%c%*[A-Z]_NODE", &type);
    switch (type)
    {
    case 'D':
	return readDecisionNode(file, tree, maxp, maxu);
    case 'C':
	return readChanceNode(file, tree, maxp, maxu);
    case 'T':
	return readTerminalNode(file, tree, maxp, maxu);
    }

    return NULL;
}

static Node readDecisionNode(FILE *file, DecisionTree tree,
			     int *maxp, int *maxu)
{
    Node node;
    int n, i;

    fscanf(file, "%d", &n);
    node = createDecisionNode(tree, n);
    for (i = 0; i < n; i++)
	node->data.d->alternatives[i] = readNode(file, tree, maxp, maxu);

    return node;
}

static Node readChanceNode(FILE *file, DecisionTree tree,
			   int *maxp, int *maxu)
{
    Node node;
    int n, i, p;

    fscanf(file, "%d", &n);
    node = createChanceNode(tree, n);
    for (i = 0; i < n; i++)
    {
	fscanf(file, "%*[\n\t ]");
	fscanf(file, "p%d", &p);
	if (p > *maxp)
	    *maxp = p;
	node->data.c->probindexes[i] = p - 1;
	node->data.c->outcomes[i] = readNode(file, tree, maxp, maxu);
    }

    return node;
}

static Node readTerminalNode(FILE *file, DecisionTree tree,
			     int *maxp, int *maxu)
{
    Node node;
    int u;

    node = createTerminalNode(tree);
    fscanf(file, "%*[\n\t ]");
    fscanf(file, "u%d", &u);
    if (u > *maxu)
	*maxu = u;
    node->data.t->utilindex = u - 1;

    return node;
}

static double expectedUtility(Node node)
{
    switch (node->type)
    {
    case DECISION_NODE:
	return expectedUtilityD(node->data.d);
    case CHANCE_NODE:
	return expectedUtilityC(node->data.c);
    case TERMINAL_NODE:
	return expectedUtilityT(node->data.t);
    }

    return 0.0; /* should never reach this point */
}

static double expectedUtilityD(DecisionNode node)
{
    double eu = expectedUtility(node->alternatives[0]);
    int i;

    for (i = 1; i < node->nalternatives; i++)
    {
	double tmp_eu = expectedUtility(node->alternatives[i]);
	if (tmp_eu > eu)
	    eu = tmp_eu;
    }

    return eu;
}

static double expectedUtilityC(ChanceNode node)
{
    double eu = node->tree->values[node->probindexes[0]] *
	expectedUtility(node->outcomes[0]);
    int i;

    for (i = 1; i < node->noutcomes; i++)
    {
	eu += node->tree->values[node->probindexes[i]] *
	    expectedUtility(node->outcomes[i]);
    }

    return eu;
}

static double expectedUtilityT(TerminalNode node)
{
    return node->tree->utilities[node->utilindex];
}

static int evaluateDecisionNode(DecisionNode node)
{
    double eu = expectedUtility(node->alternatives[0]);
    int best = 0;
#ifdef RANDOM_TIEBREAKS
    int nequal = 1;
#endif
    int i;

    for (i = 1; i < node->nalternatives; i++)
    {
	double tmp_eu = expectedUtility(node->alternatives[i]);
	if (tmp_eu > eu)
	{
	    eu = tmp_eu;
#ifdef RANDOM_TIEBREAKS
	    nequal = 1;
#endif
	    best = i;
	}
#ifdef RANDOM_TIEBREAKS
	else if (tmp_eu == eu)
	{
	    nequal++;
	    if (drand48() < 1.0 / nequal)
		best = i;
	}
#endif
    }

    return best;
}


DecisionTree readTree(const char *path)
{
    DecisionTree tree;
    FILE *file;
    int maxp = 0, maxu = 0;

    file = fopen(path, "r");
    if (file == NULL)
	return NULL;

    tree = createDecisionTree();
    tree->root = readNode(file, tree, &maxp, &maxu);
    fclose(file);
    initDecisionTree(tree, maxp, maxu);

    return tree;
}

void freeTree(DecisionTree tree)
{
    freeNode(tree->root);
    free(tree->values);
    free(tree);
}

void setValues(DecisionTree tree,
	       int first, int n, const double *values)
{
#if USE_MEMCPY
    memcpy(tree->values + first, values, n * sizeof(*values));
#else
    int i;

    for (i = 0; i < n; i++)
	tree->values[first + i] = values[i];
#endif
}

int evaluateTree(DecisionTree tree)
{
    if (tree->root->type != DECISION_NODE)
	return -1;

    return evaluateDecisionNode(tree->root->data.d);
}

#ifdef TEST
void printNode(Node node, int depth);

void printDecisionNode(DecisionNode node, int depth)
{
    int i, j;

    printf("DECISION_NODE %d\n", node->nalternatives);
    for (i = 0; i < node->nalternatives; i++)
    {
	for (j = 0; j <= depth; j++)
	    putchar('\t');
	printNode(node->alternatives[i], depth + 1);
    }
}

void printChanceNode(ChanceNode node, int depth)
{
    int i, j;

    printf("CHANCE_NODE %d\n", node->noutcomes);
    for (i = 0; i < node->noutcomes; i++)
    {
	for (j = 0; j <= depth; j++)
	    putchar('\t');
	printf("p%d ", node->probindexes[i]);
	printNode(node->outcomes[i], depth + 1);
    }
}

void printTerminalNode(TerminalNode node, int depth)
{
    printf("TERMINAL_NODE u%d\n", node->utilindex);
}

void printNode(Node node, int depth)
{
    switch (node->type)
    {
    case DECISION_NODE:
	printDecisionNode(node->data.d, depth);
	break;
    case CHANCE_NODE:
	printChanceNode(node->data.c, depth);
	break;
    case TERMINAL_NODE:
	printTerminalNode(node->data.t, depth);
	break;
    }
}

void printTree(DecisionTree tree)
{
    printNode(tree->root, 0);
}

static double rand01(void)
{
    return rand() / (RAND_MAX + 1.0);
}

void get_probabilities(int p, int s, double *probs)
{
    while (p--)
    {
        int i = s - 1;
        double last = 1.0;

        while (i--)
            last -= (*probs++ = last * rand01());
        *probs++ = last;
    }
}

void get_utilities(int p, double *utils)
{
    while (p--)
        *utils++ = rand01();
}

int main(int argc, char **argv)
{
    DecisionTree tree;
    int i, ntrials = 100;
    long seed = time(NULL);
    int advice;
    double m;
    double sum = 0.0;
    double sqsum = 0.0;
    static struct timeval start, stop;
    double values[100];

    if (argc > 1)
	ntrials = atoi(argv[1]);
    if (argc > 2)
	seed = atoi(argv[2]);

    tree = readTree("hasball2.dt");
    for (i = 0; i < ntrials; i++)
    {
	gettimeofday(&start, NULL);
	get_probabilities(2, 2, values);
	get_probabilities(12, 4, values + 4);
	get_utilities(12, values + 52);
	setValues(tree, 0, 64, values);
	advice = evaluateTree(tree);
	gettimeofday(&stop, NULL);
        if (stop.tv_usec < start.tv_usec)
        {
            stop.tv_usec += 1000000;
            start.tv_sec += 1;
        }
        m = 1000.0 * (stop.tv_sec - start.tv_sec) +
            (stop.tv_usec - start.tv_usec) / 1000.0;
#if 0
        printf("%.3f (%d)\n", m, advice);
#endif
        sum += m;
        sqsum += m * m;
    }
    printf("mean is %f\n", sum / ntrials);
    printf("deviation is %f\n", sqrt((sqsum - sum * sum / ntrials) / ntrials));

    freeTree(tree);

    return 0;
}
#endif
