package decisionanalysis;

import java.io.StreamTokenizer;
import java.io.IOException;

/**
 * A decision tree chance node.
 *
 * @author H&aring;kan L. Younes
 */
final class DTChanceNode extends DTNode
{
    /**
     * Number of probability variables that have been used.
     */
    protected static int maxP = 0;


    /**
     * The descendants of this node.
     */
    private DTNode[] outcomes;

    /**
     * Indexes to probabilities of the outcomes of this node in the
     * value array of the tree that this node belongs to.
     */
    private int[] probIndexes;


    /**
     * Reads a chance node from the given source.
     *
     * @param st
     *  source for node description.
     * @param tree
     *  decision tree that the node belongs to.
     * @return the read node.
     * @exception IOException
     *  if an I/O error occurs.
     */
    protected static DTNode readFrom(StreamTokenizer st, DecisionTree tree)
	 throws IOException
    {
	if (st.nextToken() != StreamTokenizer.TT_NUMBER)
	{
	    throw new IOException("illegal token: " + st);
	}
	int n = (int)st.nval;
	DTChanceNode node = new DTChanceNode(tree, n);
	for (int i = 0; i < n; i++)
	{
	    if (st.nextToken() != 'p')
	    {
		throw new IOException("illegal token: " + st);
	    }
	    if (st.nextToken() != StreamTokenizer.TT_NUMBER)
	    {
		throw new IOException("illegal token: " + st);
	    }
	    int p = (int)st.nval;
	    if (p > maxP)
	    {
		maxP = p;
	    }
	    node.probIndexes[i] = p - 1;
	    node.outcomes[i] = DTNode.readFrom(st, tree);
	}
	return node;
    }


    /**
     * Constructs a decision tree chance node.
     *
     * @param tree
     *  decision tree that the node belongs to.
     * @param numOutcomes
     *  number of possible outcomes.
     */
    public DTChanceNode(DecisionTree tree, int numOutcomes)
    {
	super(tree);
	outcomes = new DTNode[numOutcomes];
	probIndexes = new int[numOutcomes];
    }


    /**
     * Returns the expected utility for this node.
     *
     * @return the expected utility for this node.
     */
    public double expectedUtility()
    {
	double eu = tree.getProbability(probIndexes[0]) *
	    outcomes[0].expectedUtility();
	for (int i = 1; i < outcomes.length; i++)
	{
	    eu += tree.getProbability(probIndexes[i]) *
		outcomes[i].expectedUtility();
	}
	return eu;
    }

    /**
     * Returns a string representation of this node.
     *
     * @param depth
     *  depth of node in decision tree.
     * @return a string representation of this node.
     */
    public String toString(int depth)
    {
	StringBuffer sb = new StringBuffer("CHANCE_NODE ");
	sb.append(outcomes.length);
	for (int i = 0; i < outcomes.length; i++)
	{
	    sb.append('\n');
	    for (int j = 0; j <= depth; j++)
	    {
		sb.append('\t');
	    }
	    sb.append('p').append(probIndexes[i] + 1).append(' ');
	    sb.append(outcomes[i].toString(depth + 1));
	}
	return sb.toString();
    }
}
