package decisionanalysis;

import java.io.Serializable;
import java.io.FileReader;
import java.io.StreamTokenizer;
import java.io.FileNotFoundException;
import java.io.IOException;

/**
 * A decision tree.
 *
 * @author H&aring;kan L. Younes
 */
public final class DecisionTree extends Object implements Serializable
{
    /**
     * The root node.
     */
    private DTDecisionNode root;

    /**
     * Array of values with probabilities first, followed by utilities.
     */
    private double[] values;

    /**
     * Index of first utility in <code>values</code>.
     */
    private int utilBase;


    /**
     * Constructs a decision tree from a text file.
     *
     * @param fileName
     *  name of file containing description of decision tree.
     * @exception FileNotFoundException
     *  if the given file could not be found.
     * @exception IOException
     *  if an I/O error occurs.
     */
    public DecisionTree(String fileName)
	 throws FileNotFoundException, IOException
    {
	FileReader reader = new FileReader(fileName);
	StreamTokenizer st = new StreamTokenizer(reader);
	st.wordChars('_', '_');
	st.ordinaryChar('p');
	st.ordinaryChar('u');
	root = (DTDecisionNode)DTNode.readFrom(st, this);
	values = new double[DTChanceNode.maxP + DTTerminalNode.maxU];
	utilBase = DTChanceNode.maxP;
    }


    /**
     * Returns the probability with the given index.
     *
     * @param index
     *  an index.
     * @return the probability with the given index.
     */
    public double getProbability(int index)
    {
	return values[index];
    }

    /**
     * Returns the utility with the given index.
     *
     * @param index
     *  an index.
     * @return the utility with the given index.
     */
    public double getUtility(int index)
    {
	return values[utilBase + index];
    }

    /**
     * Sets values in the decision tree.
     *
     * @param values
     *  new values.
     * @param first
     *  index of first value to set.
     */
    public void setValues(double[] values, int first)
    {
	System.arraycopy(values, 0, this.values, first, values.length);
    }

    /**
     * Returns the decision that maximizes the expected utility.
     *
     * @return the decision that maximizes the expected utility.
     */
    public int getOptimalDecision()
    {
	return root.getOptimalDecision();
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a string representation of this object.
     */
    public String toString()
    {
	return root.toString(0);
    }
}
