/**
 * smile_pronouncer.h
 * ==================
 * Interface to basic pronouncer functionality based on SMILE.
 *
 * By Hkan L. Younes (lorens@acm.org)
 */

#ifndef SMILE_PRONOUNCER_H
#define SMILE_PRONOUNCER_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * Initialize pronouncer.
 *
 * Returns -1 on failure, and 0 on success.
 *
 * N.B. must be called before the pronouncer is used.
 */
extern int initializePronouncer(void);


/**
 * Finalizes pronouncer.
 *
 * N.B. should be called after the last call to the pronouncer.
 */
extern void finalizePronouncer(void);


/**
 * Loads a model.
 *
 * Parameters:
 *  path  - path to file containing the model.
 *
 * Returns -1 on failure, and an identifier for the model on success.
 */
extern int loadModel(char *path);


/**
 * Frees a model.
 *
 * Parameters:
 *  modelId - a model identifier as returned by loadModel.
 */
extern void freeModel(int modelId);


/**
 * Sets probabilities in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating row in probability table.
 *  nprobs       - number of probabilities in probs.
 *  probs        - array of probabilities to assign to row.
 */
extern void setProbabilities(int modelId, char *nodeName,
			     const int *parentStates, int nprobs,
			     const double *probs);


/**
 * Sets a utility in the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  nodeName     - name of node.
 *  parentStates - states of parents indicating position in utility table.
 *  util         - the utility.
 */
extern void setUtility(int modelId, char *node_name,
		       const int *parent_states, double util);


/**
 * Sets all probabilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  nstates       - number of states that this node can be in.
 *  probs         - probabilities.
 */
extern void setAllProbabilities(int modelId, char *nodeName,
				int nparents, const int *parentNstates,
				int nstates, const double *probs);


/**
 * Sets all utilities in a node of the given model.
 *
 * Parameters:
 *  modelId       - a model identifier as returned by loadModel.
 *  nodeName      - name of node.
 *  nparents      - number of parents
 *  parentNstates - number of states that each of the parents can be in.
 *  utils         - utilities.
 */
extern void setAllUtilities(int modelId, char *nodeName, 
			    int nparents, const int *parentNstates,
			    const double *utils);


/**
 * Evaluates the given model.
 *
 * Parameters:
 *  modelId      - a model identifier as returned by loadModel.
 *  decisionName - name of decision node.
 *
 * Returns the index of an alternative that maximizes the expected utility.
 */
extern int evaluateModel(int modelId, char *decision_name);

#ifdef __cplusplus
}
#endif

#endif /* SMILE_PRONOUNCER_H */
